"""
Container component for composing multiple components, such as Sequential
and ComponentList.

This design draws inspiration from PyTorch’s modular
container patterns, including `nn.Sequential` and `nn.ModuleList`. The
`Container` component allows for grouping several components into one, enabling
flexible and reusable model architectures.

Design Motivation:
-------------------
This implementation follows the same principles as PyTorch’s component-based
design, encouraging modularity, reusability, and extensibility. The `Container`
component provides an easy way to manage multiple layers or other components,
while ensuring that their parameters are properly registered and updated during
training.

Credits:
---------
The design of this component takes inspiration from the PyTorch project
(https://pytorch.org). PyTorch is an open-source deep learning framework,
licensed under a BSD-style license. Although this code is not part of the
official PyTorch library, it mirrors the same design principles.

For more details on PyTorch’s licensing, refer to:
https://github.com/pytorch/pytorch/blob/main/LICENSE

Usage Example:
--------------
    class MyModule(nn.Module):
        def __init__(self):
            super().__init__()

            self.model = nn.Sequential(
                  nn.Conv2d(1,20,5),
                  nn.ReLU(),
                  nn.Conv2d(20,64,5),
                  nn.ReLU()
                )
            self.linears = nn.ModuleList([nn.Linear(10, 10) for i in range(10)])

        def forward(self, x):
            # ModuleList can act as an iterable, or be indexed using ints
            for i, l in enumerate(self.linears):
                x = self.linears[i // 2](x) + l(x)
            return x

"""

from collections import OrderedDict, abc as container_abcs
import operator
from itertools import islice, chain
from typing import TypeVar, Dict, Union, Iterable, Iterator, Any, overload, Optional

from adalflow.core.component import Component

T = TypeVar("T", bound=Component)

__all__ = ["Sequential", "ComponentList"]


class Sequential(Component):
    __doc__ = r"""A sequential container.

    Adapted from PyTorch's ``nn.Sequential``.

    Components will be added to it in the order they are passed to the constructor.
    Alternatively, an ``OrderedDict`` of components can be passed in.
    It "chains" outputs of the previous component to the input of the next component sequentially.
    Output of the previous component is input to the next component as positional argument.

    Benefits of using Sequential:
    1. Convenient for data pipeline that often consists of multiple components. This allow users to encapsulate the pipeline in a single component.
    Examples:

    Without Sequential:

    .. code-block:: python

        class AddAB(Component):
            def call(self, a: int, b: int) -> int:
                return a + b


        class MultiplyByTwo(Component):
            def call(self, input: int) -> int:
                return input * 2

        class DivideByThree(Component):
            def call(self, input: int) -> int:
                return input / 3

        # Manually chaining the components
        add_a_b = AddAB()
        multiply_by_two = MultiplyByTwo()
        divide_by_three = DivideByThree()

        result = divide_by_three(multiply_by_two(add_a_b(2, 3)))



    With Sequential:

    .. code-block:: python

        seq = Sequential(AddAB(), MultiplyByTwo(), DivideByThree())
        result = seq(2, 3)

    .. note::
        Only the first component can receive arbitrary positional and keyword arguments.
        The rest of the components should have a single positional argument as input and have it to be exactly the same type as the output of the previous component.

    2. Apply a transformation or operation (like training, evaluation, or serialization) to the Sequential object, it automatically applies that operation to each component it contains.
    This can be useful for In-context learning training.


    Examples:

    1. Use positional arguments:
        >>> seq = Sequential(component1, component2)
    2. Add components:
        >>> seq.append(component4)
    3. Get a component:
        >>> seq[0]
    4. Delete a component:
        >>> del seq[0]
    5. Iterate over components:
        >>> for component in seq:
        >>>     print(component)
    6. Add two Sequentials:
        >>> seq1 = Sequential(component1, component2)
        >>> seq2 = Sequential(component3, component4)
        >>> seq3 = seq1 + seq2
    7. Use OrderedDict:
        >>> seq = Sequential(OrderedDict({"component1": component1, "component2": component2}))
    8. Index OrderDict:
        >>> seq = Sequential(OrderedDict({"component1": component1, "component2": component2}))
        >>> seq["component1"]
        # or
        >>> seq[0]
    9. Call with a single argument as input:
        >>> seq = Sequential(component1, component2)
        >>> result = seq.call(2)
    10. Call with multiple arguments as input:
        >>> seq = Sequential(component1, component2)
        >>> result = seq.call(2, 3)
    """

    _components: Dict[str, Component] = OrderedDict()  # type: ignore[assignment]

    @overload
    def __init__(self, *args: Component) -> None: ...

    @overload
    def __init__(self, arg: "OrderedDict[str, Component]") -> None: ...

    def __init__(self, *args):
        super().__init__()
        if len(args) == 1 and isinstance(args[0], OrderedDict):
            for key, component in args[0].items():
                self.add_component(key, component)
        else:
            for idx, component in enumerate(args):
                self.add_component(str(idx), component)

    def _get_item_by_idx(self, iterator: Iterator[Component], idx: int) -> Component:
        """Get the idx-th item of the iterator."""
        size = len(self)
        idx = operator.index(idx)
        if not -size <= idx < size:
            raise IndexError(f"index {idx} is out of range")
        idx %= size
        return next(islice(iterator, idx, None))

    def __getitem__(
        self, idx: Union[slice, int, str]
    ) -> Union["Sequential", Component]:
        """Get the idx-th and by-key component of the Sequential."""
        if isinstance(idx, slice):
            return self.__class__(OrderedDict(list(self._components.items())[idx]))
        elif isinstance(idx, str):
            return self._components[idx]
        else:
            return self._get_item_by_idx(iter(self._components.values()), idx)

    def __setitem__(self, idx: Union[int, str], component: Component) -> None:
        """Set the idx-th component of the Sequential."""
        if isinstance(idx, str):
            self._components[idx] = component
        else:
            # key: str = self._get_item_by_idx(iter(self._components.keys()), idx)
            # self._components[key] = component
            key_list = list(self._components.keys())
            key = key_list[idx]
            self._components[key] = component

    def __delitem__(self, idx: Union[slice, int, str]) -> None:
        """Delete the idx-th component of the Sequential."""
        if isinstance(idx, slice):
            for key in list(self._components.keys())[idx]:
                delattr(self, key)
        elif isinstance(idx, str):
            del self._components[idx]
        else:
            # key = self._get_item_by_idx(iter(self._components.keys()), idx)
            key_list = list(self._components.keys())
            key = key_list[idx]

            delattr(self, key)

        # Reordering is needed if numerical keys are used to keep the sequence
        self._components = OrderedDict(
            (str(i), comp) for i, comp in enumerate(self._components.values())
        )

    def __iter__(self) -> Iterator[Component]:
        r"""Iterates over the components of the Sequential.

        Examples:
        1. Iterate over the components:

        .. code-block:: python

            for component in seq:
                print(component)
        """
        return iter(self._components.values())

    def __len__(self) -> int:
        return len(self._components)

    def __add__(self, other) -> "Sequential":
        r"""Adds two Sequentials.

        Creating a new Sequential with components of both the Sequentials.

        Examples:
        1. Add two Sequentials:

        .. code-block:: python

            seq1 = Sequential(component1, component2)
            seq2 = Sequential(component3, component4)
            seq3 = seq1 + seq2
        """
        if isinstance(other, Sequential):
            ret = Sequential()
            for layer in self:
                ret.append(layer)
            for layer in other:
                ret.append(layer)
            return ret
        else:
            raise ValueError(
                "add operator supports only objects "
                f"of Sequential class, but {str(type(other))} is given."
            )

    def __iadd__(self, other) -> "Sequential":
        r"""Inplace add two Sequentials.

        Adding components of the other Sequential to the current Sequential.

        Examples:
        1. Inplace add two Sequentials:

        .. code-block:: python

            seq1 = Sequential(component1, component2)
            seq2 = Sequential(component3, component4)
            seq1 += seq2
        """
        if not isinstance(other, Sequential):
            raise ValueError(
                "add operator supports only objects "
                f"of Sequential class, but {str(type(other))} is given."
            )
        for layer in other:
            self.append(layer)
        return self

    @overload
    def call(self, input: Any) -> object: ...

    @overload
    def call(self, *args: Any, **kwargs: Any) -> object: ...

    def call(self, *args: Any, **kwargs: Any) -> object:
        if len(args) == 1 and not kwargs:
            input = args[0]
            for component in self._components.values():
                input = component(input)
            return input
        else:
            for component in self._components.values():
                result = component(*args, **kwargs)
                if (
                    isinstance(result, tuple)
                    and len(result) == 2
                    and isinstance(result[1], dict)
                ):
                    args, kwargs = result
                else:
                    args = (result,)
                    kwargs = {}
            return args[0] if len(args) == 1 else (args, kwargs)

    @overload
    async def acall(self, input: Any) -> object: ...

    @overload
    async def acall(self, *args: Any, **kwargs: Any) -> object: ...

    async def acall(self, *args: Any, **kwargs: Any) -> object:
        r"""When you for loop or multiple await calls inside each component, use acall method can potentially speed up the execution."""
        if len(args) == 1 and not kwargs:
            input = args[0]
            for component in self._components.values():
                input = await component(input)
            return input
        else:
            for component in self._components.values():
                result = await component(*args, **kwargs)
                if (
                    isinstance(result, tuple)
                    and len(result) == 2
                    and isinstance(result[1], dict)
                ):
                    args, kwargs = result
                else:
                    args = (result,)
                    kwargs = {}
            return args[0] if len(args) == 1 else (args, kwargs)

    def append(self, component: Component) -> "Sequential":
        r"""Appends a component to the end of the Sequential."""
        idx = len(self._components)
        self.add_component(str(idx), component)
        return self

    def insert(self, idx: int, component: Component) -> None:
        r"""Inserts a component at a given index in the Sequential."""
        if not isinstance(component, Component):
            raise TypeError(
                f"component should be an instance of Component, but got {type(component)}"
            )
        n = len(self._components)
        if not (-n <= idx <= n):
            raise IndexError(
                f"index {idx} is out of range for Sequential with length {len(self._components)}"
            )
        if idx < 0:
            idx += n
        for i in range(n, idx, -1):
            self._components[str(i)] = self._components[str(i - 1)]
        self._components[str(idx)] = component

    def extend(self, components: Iterable[Component]) -> "Sequential":
        r"""Extends the Sequential with components from an iterable."""
        for component in components:
            self.append(component)
        return self


def _addindent(s_: str, numSpaces: int):
    s = s_.split("\n")
    # don't do anything for single-line stuff
    if len(s) == 1:
        return s_
    first = s.pop(0)
    s = [(numSpaces * " ") + line for line in s]
    s = "\n".join(s)
    s = first + "\n" + s
    return s


class ComponentList(Component):
    __doc__ = r"""Holds subcomponents in a list.

    :class:`adalflow.core.ComponentList` can be indexed like a regular Python list, but
    the components it holds are properly registered, and will be visible by all
    :class:`adalflow.core.Component` methods.

    Args:
        components (iterable, optional): an iterable of components to add

    Examples:

    .. code-block:: python

        # Example of how to use ComponentList
        class MyComponents(Component):
            def __init__(self):
                super().__init__()
                self.llms = ComponentList([adal.Generator() for i in range(10)])

            def forward(self, x):
                for layer in self.layers:
                    x = layer(x)
                return x
    """
    _components: Dict[str, Component] = OrderedDict()

    def __init__(self, components: Optional[Iterable[Component]] = None) -> None:
        super().__init__()
        if components is not None:
            self += components

    def _get_abs_string_index(self, idx):
        """Get the absolute index as a string."""
        idx = operator.index(idx)
        if not (-len(self) <= idx < len(self)):
            raise IndexError(f"index {idx} is out of range")
        if idx < 0:
            idx += len(self)
        return str(idx)

    def __getitem__(self, idx: Union[int, slice]) -> Union[Component, "ComponentList"]:
        """Retrieve a component or a slice of components."""
        if isinstance(idx, slice):
            return self.__class__(list(self._components.values())[idx])
        else:
            return self._components[self._get_abs_string_index(idx)]

    def __setitem__(self, idx: int, component: Component) -> None:
        """Set a component at the given index."""
        idx = self._get_abs_string_index(idx)
        return setattr(self, str(idx), component)

    def __delitem__(self, idx: Union[int, slice]) -> None:
        """Delete a component or a slice of components."""
        if isinstance(idx, slice):
            for k in range(len(self._components))[idx]:
                delattr(self, str(k))
        else:
            delattr(self, self._get_abs_string_index(idx))
        # To preserve numbering, self._components is being reconstructed with modules after deletion
        str_indices = [str(i) for i in range(len(self._components))]
        self._components = OrderedDict(
            list(zip(str_indices, self._components.values()))
        )

    def __len__(self) -> int:
        """Return the number of components."""
        return len(self._components)

    def __iter__(self) -> Iterator[Component]:
        """Iterate over the components."""
        return iter(self._components.values())

    def __iadd__(self, components: Iterable[Component]) -> "ComponentList":
        """Add multiple components using the `+=` operator."""

        return self.extend(components)

    def __add__(self, other: Iterable[Component]) -> "ComponentList":
        """Concatenate two ComponentLists."""

        combined = ComponentList()
        for i, component in enumerate(chain(self, other)):
            combined.add_component(str(i), component)
        return combined

    def __repr__(self):
        """Return a custom repr for ModuleList that compresses repeated module representations."""
        list_of_reprs = [repr(item) for item in self]
        if len(list_of_reprs) == 0:
            return self._get_name() + "()"

        start_end_indices = [[0, 0]]
        repeated_blocks = [list_of_reprs[0]]
        for i, r in enumerate(list_of_reprs[1:], 1):
            if r == repeated_blocks[-1]:
                start_end_indices[-1][1] += 1
                continue

            start_end_indices.append([i, i])
            repeated_blocks.append(r)

        lines = []
        main_str = self._get_name() + "("
        for (start_id, end_id), b in zip(start_end_indices, repeated_blocks):
            local_repr = f"({start_id}): {b}"  # default repr

            if start_id != end_id:
                n = end_id - start_id + 1
                local_repr = f"({start_id}-{end_id}): {n} x {b}"

            local_repr = _addindent(local_repr, 2)
            lines.append(local_repr)

        main_str += "\n  " + "\n  ".join(lines) + "\n"
        main_str += ")"
        return main_str

    def __dir__(self):
        keys = super().__dir__()
        keys = [key for key in keys if not key.isdigit()]
        return keys

    def insert(self, index: int, component: Component) -> None:
        """Insert a component at the specified index."""
        for i in range(len(self._components), index, -1):
            self._components[str(i)] = self._components[str(i - 1)]
        self._components[str(index)] = component

    def pop(self, index: Union[int, slice]) -> Component:
        """Remove and return a component at the given index."""
        component = self[index]
        del self[index]
        return component

    def append(self, component: Component) -> "ComponentList":
        """Append a component to the list."""
        # self._components[str(len(self))] = component
        self.add_component(str(len(self)), component)
        return self

    def extend(self, components: Iterable[Component]) -> "ComponentList":
        """Extend the list by appending multiple components."""
        # for component in components:
        #     self.append(component)
        # return self

        if not isinstance(components, container_abcs.Iterable):
            raise TypeError(
                "ModuleList.extend should be called with an "
                "iterable, but got " + type(components).__name__
            )
        offset = len(self)
        for i, component in enumerate(components):
            self.add_component(str(offset + i), component)
        return self


# TODO: need to do the same to ParameterList and ParameterDict, ModuleDict



"""DataClassParser will help users interact with LLMs even better than JsonOutputParser and YamlOutputParser with DataClass."""

from dataclasses import is_dataclass
from typing import Any, Literal, List, Optional
import logging

from adalflow.core.prompt_builder import Prompt
from adalflow.core.component import DataComponent
from adalflow.core.string_parser import YamlParser, JsonParser
from adalflow.core.base_data_class import DataClass, DataClassFormatType
from adalflow.core.base_data_class import ExcludeType, IncludeType


__all__ = ["DataClassParser"]

log = logging.getLogger(__name__)

JSON_OUTPUT_FORMAT = r"""Your output should be formatted as a standard JSON instance with the following schema:
```
{{schema}}
```
-Make sure to always enclose the JSON output in triple backticks (```). Please do not add anything other than valid JSON output!
-Use double quotes for the keys and string values.
-DO NOT mistaken the "properties" and "type" in the schema as the actual fields in the JSON output.
-Follow the JSON formatting conventions."""

YAML_OUTPUT_FORMAT = r"""Your output should be formatted as a standard YAML instance with the following schema:
```
{{schema}}
```
-Make sure to always enclose the YAML output in triple backticks (```). Please do not add anything other than valid YAML output!
-Follow the YAML formatting conventions with an indent of 2 spaces.
-DO NOT mistaken the "properties" and "type" in the schema as the actual fields in the YAML output.
-Quote the string values properly."""

EXAMPLES_FORMAT = r"""
{% if examples %}
{% for example in examples %}
{{example}}
__________
{% endfor %}
{% endif %}
"""


class DataClassParser(DataComponent):
    __doc__ = r"""Made the structured output even simpler compared with JsonOutputParser and YamlOutputParser.

        1. Understands __input_fields__ and __output_fields__ from the DataClass (no need to use include/exclude to decide fields).
        2. User can choose to save the `task_desc` in the DataClass and use it in the prompt.

        Example:

        .. code-block:: python

            @dataclass
            class BasicQAOutput(adal.DataClass):
                explanation: str = field(
                    metadata={"desc": "A brief explanation of the concept in one sentence."}
                )
                example: str = field(
                    metadata={"desc": "An example of the concept in a sentence."}
                )
                # Control output fields order
                __output_fields__ = ["explanation", "example"]

            # Define the template using jinja2 syntax
            qa_template = "<SYS>
            You are a helpful assistant.
            <OUTPUT_FORMAT>
            {{output_format_str}}
            </OUTPUT_FORMAT>
            </SYS>
            <USER> {{input_str}} </USER>"

            parser = adal.DataClassParser(data_class=BasicQAOutput, return_data_class=True)

            # Set up the generator with model, template, and parser
            self.generator = adal.Generator(
                model_client=model_client,
                model_kwargs=model_kwargs,
                template=qa_template,
                prompt_kwargs={"output_format_str": parser.get_output_format_str()},
                output_processors=parser,
            )
        """

    def __init__(
        self,
        data_class: DataClass,
        return_data_class: bool = False,
        format_type: Literal["yaml", "json"] = "json",
    ):
        super().__init__()
        if not is_dataclass(data_class):
            raise ValueError("data_class must be a dataclass.")

        if not issubclass(data_class, DataClass):
            raise ValueError("data_class must be a subclass of DataClass.")

        self._return_data_class = return_data_class
        self._input_fields = data_class.get_input_fields()
        self._output_fields = data_class.get_output_fields()
        if format_type not in ["yaml", "json"]:
            raise ValueError("Invalid format type.")
        self._format_type = format_type
        self._data_class: DataClass = data_class
        self._output_processor = YamlParser() if format_type == "yaml" else JsonParser()
        self.output_format_prompt = (
            Prompt(template=YAML_OUTPUT_FORMAT)
            if format_type == "yaml"
            else Prompt(template=JSON_OUTPUT_FORMAT)
        )

    def get_input_format_str(self) -> str:
        r"""Return the formatted instructions to use in prompt for the input format."""
        if self._format_type == "yaml":
            return self._data_class.to_yaml_signature(include=self._input_fields)
        else:
            return self._data_class.to_json_signature(include=self._input_fields)

    def get_output_format_str(self) -> str:
        r"""Return the formatted instructions to use in prompt for the output format."""
        output_format_str = None
        if self._format_type == "yaml":
            schema = self._data_class.to_yaml_signature(include=self._output_fields)
            output_format_str = Prompt(template=YAML_OUTPUT_FORMAT)(schema=schema)
        else:
            schema = self._data_class.to_json_signature(include=self._output_fields)
            output_format_str = Prompt(template=JSON_OUTPUT_FORMAT)(schema=schema)
        return output_format_str

    def get_input_str(self, input: DataClass) -> str:
        r"""Return the formatted input string."""
        if not isinstance(input, self._data_class):
            raise ValueError("input must be an instance of the data_class.")
        if self._format_type == "yaml":
            return input.to_yaml(include=self._input_fields)
        else:
            return input.to_json(include=self._input_fields)

    def get_task_desc_str(self) -> str:
        r"""Return the task description string."""
        return self._data_class.get_task_desc()

    def get_examples_str(
        self,
        examples: List[DataClass],
        include: Optional[IncludeType] = None,
        exclude: Optional[ExcludeType] = None,
    ) -> str:
        r"""Return the examples string."""
        str_examples = []
        if examples and len(examples) > 0:
            for example in examples:
                per_example_str = example.format_example_str(
                    format_type=(
                        DataClassFormatType.EXAMPLE_YAML
                        if self._format_type == "yaml"
                        else DataClassFormatType.EXAMPLE_JSON
                    ),
                    exclude=exclude,
                    include=include,
                )
                str_examples.append(per_example_str)

        examples_str = Prompt(template=EXAMPLES_FORMAT)(examples=str_examples)
        return examples_str

    def call(self, input: str) -> Any:
        r"""Parse the output string to the desired format and return the parsed output."""
        try:
            output_dict = self._output_processor(input)
            if self._return_data_class:
                return self._data_class.from_dict(output_dict)
            return output_dict
        except Exception as e:
            log.error(f"Error at parsing output: {e}")
            raise ValueError(f"Error: {e}")

    def _extra_repr(self) -> str:
        s = f"data_class={self._data_class.__name__}, format_type={self._format_type},\
            return_data_class={self._return_data_class}, input_fields={self._input_fields},\
            output_fields={self._output_fields}"
        return s



from .big_bench_hard import BigBenchHard
from .hotpot_qa import HotPotQA
from .trec import TrecDataset
from .types import Example, HotPotQAData, TrecData

__all__ = [
    "BigBenchHard",
    "HotPotQA",
    "Example",
    "HotPotQAData",
    "TrecDataset",
    "TrecData",
]



"""Helper components for data transformation such as embeddings and document splitting."""

from copy import deepcopy
from typing import List, TypeVar, Sequence, Union, Dict, Any
from tqdm import tqdm


from adalflow.core.component import Component

from adalflow.core.types import (
    Document,
    RetrieverOutput,
)
from adalflow.core.embedder import (
    BatchEmbedder,
    BatchEmbedderOutputType,
    BatchEmbedderInputType,
    Embedder,
)


T = TypeVar("T")
__all__ = [
    "ToEmbeddings",
    "RetrieverOutputToContextStr",
    "retriever_output_to_context_str",
]

# TODO: make the GeneratorOutput include the token usage too.


def retriever_output_to_context_str(
    retriever_output: Union[RetrieverOutput, List[RetrieverOutput]],
    deduplicate: bool = False,
) -> str:
    r"""The retrieved documents from one or multiple queries.
    Deduplicate is especially helpful when you used query expansion.
    """
    """
    How to combine your retrieved chunks into the context is highly dependent on your use case.
    If you used query expansion, you might want to deduplicate the chunks.
    """
    chunks_to_use: List[Document] = []
    context_str = ""
    sep = " "
    if isinstance(retriever_output, RetrieverOutput):
        chunks_to_use = retriever_output.documents
    else:
        for output in retriever_output:
            chunks_to_use.extend(output.documents)
    if deduplicate:
        unique_chunks_ids = set([chunk.id for chunk in chunks_to_use])
        # id and if it is used, it will be True
        used_chunk_in_context_str: Dict[Any, bool] = {
            id: False for id in unique_chunks_ids
        }
        for chunk in chunks_to_use:
            if not used_chunk_in_context_str[chunk.id]:
                context_str += sep + chunk.text
                used_chunk_in_context_str[chunk.id] = True
    else:
        context_str = sep.join([chunk.text for chunk in chunks_to_use])
    return context_str


"""
For now these are the data transformation components
"""

ToEmbeddingsInputType = Sequence[Document]
ToEmbeddingsOutputType = Sequence[Document]


class ToEmbeddings(Component):
    r"""It transforms a Sequence of Chunks or Documents to a List of Embeddings.

    It operates on a copy of the input data, and does not modify the input data.
    """

    def __init__(self, embedder: Embedder, batch_size: int = 50) -> None:
        super().__init__(batch_size=batch_size)
        self.embedder = embedder
        self.batch_size = batch_size
        self.batch_embedder = BatchEmbedder(embedder=embedder, batch_size=batch_size)

    def __call__(self, input: ToEmbeddingsInputType) -> ToEmbeddingsOutputType:
        output = deepcopy(input)
        # convert documents to a list of strings
        embedder_input: BatchEmbedderInputType = [chunk.text for chunk in output]
        outputs: BatchEmbedderOutputType = self.batch_embedder(input=embedder_input)
        # n them back to the original order along with its query
        for batch_idx, batch_output in tqdm(
            enumerate(outputs), desc="Adding embeddings to documents from batch"
        ):
            for idx, embedding in enumerate(batch_output.data):
                output[batch_idx * self.batch_size + idx].vector = embedding.embedding
        return output

    def _extra_repr(self) -> str:
        s = f"batch_size={self.batch_size}"
        return s


class RetrieverOutputToContextStr(Component):
    r"""
    Wrap on functional F.retriever_output_to_context_str
    """

    def __init__(self, deduplicate: bool = False):
        super().__init__()
        self.deduplicate = deduplicate

    def __call__(
        self,
        input: Union[RetrieverOutput, List[RetrieverOutput]],
    ) -> str:
        return retriever_output_to_context_str(
            retriever_output=input, deduplicate=self.deduplicate
        )

    def _extra_repr(self) -> str:
        s = f"deduplicate={self.deduplicate}"
        return s



"""Retriever Recall @k metric."""

from typing import List, Dict

from adalflow.eval.base import BaseEvaluator, EvaluationResult
from adalflow.eval.utils import normalize_answer


class RetrieverEvaluator(BaseEvaluator):
    __doc__ = r"""Return Recall@k and Precision@k.

    Recall@k = Number of relevant retrieved documents/ Total number of relevant documents (len(gt_contexts))
    Precision@k = Number of relevant retrieved documents/ Total number of retrieved documents (len(retrieved_contexts))


    In our implementation, we use exact string matching between each gt context and the joined retrieved context string.
    You can use the longest common subsequence (LCS) or other similarity metrics(or embedding based) to decide if it is a match or not.

    You can also pass ids of retrieved and the reference.

    If you do not even have the ground truth context, but only grounth truth answers, you can consider using
    RAGAS framework for now. It computes the recall as:

    Recall = [GT statements that can be attributed to the retrieved context] / [GT statements]

    Examples:
        >>> all_retrieved_context = [
        ["Apple is founded before Google.",
        "Feburary has 28 days in common years. Feburary has 29 days in leap years. Feburary is the second month of the year.",
        ]
        >>> all_gt_context = [
            [
                "Apple is founded in 1976.",
                "Google is founded in 1998.",
                "Apple is founded before Google.",
            ],
            ["Feburary has 28 days in common years", "Feburary has 29 days in leap years"],
        ]
        >>> retriever_recall = RetrieverRecall()
        >>> avg_recall, recall_list = retriever_recall.compute(all_retrieved_context, all_gt_context)
        >>> avg_recall
        2 / 3
        >>> recall_list
        [1 / 3, 1.0]

    References:
        - RAGAS: https://docs.ragas.io/en/stable/concepts/metrics/context_recall.html
    """

    def __init__(self):
        super().__init__()

    def compute_single_item(
        self, retrieved_context: List[str], gt_context: List[str]
    ) -> Dict[str, float]:
        r"""
        Compute the recall of the retrieved context for a single query.

        Args:
            retrieved_context (List[str]): List of retrieved context strings.
            gt_context (List[str]): List of ground truth context strings.

        Returns:
            float: Recall value.
        """
        # 1 normalize the text
        normalized_retrieved_context = [
            normalize_answer(doc) for doc in retrieved_context
        ]

        normalized_gt_context = [normalize_answer(doc) for doc in gt_context]

        set_retrieved = set(normalized_retrieved_context)
        set_gt = set(normalized_gt_context)

        # 2 calculate the recall with intersection

        recall = len(set_gt.intersection(set_retrieved)) / len(set_gt)
        precision = len(set_gt.intersection(set_retrieved)) / len(set_retrieved)

        return {"recall": recall, "precision": precision}

    def compute(
        self,
        retrieved_contexts: List[List[str]],
        gt_contexts: List[List[str]],
    ) -> EvaluationResult:
        r"""
        Compute the recall of the retrieved context for a list of queries.
        Args:
            retrieved_context: List of retrieved context strings.
            gt_contexts ( List[List[str]]): List of ground truth context strings.

        Returns:
            tuple:
                - float: Average recall value.
                - List[float]: Recall values for each query.
        """
        if len(retrieved_contexts) != len(gt_contexts):
            raise ValueError(
                "The number of retrieved context lists and ground truth context lists should be the same."
            )
        k = len(retrieved_contexts[0])
        metric_list = []
        for retrieved_context, gt_context in zip(retrieved_contexts, gt_contexts):

            metric = self.compute_single_item(retrieved_context, gt_context)
            metric_list.append(metric)

        # average through each key value

        avg_recall = sum([metric["recall"] for metric in metric_list]) / len(
            metric_list
        )
        avg_precision = sum([metric["precision"] for metric in metric_list]) / len(
            metric_list
        )

        return {
            "avg_recall": avg_recall,
            "avg_precision": avg_precision,
            "recall_list": [metric["recall"] for metric in metric_list],
            "precision_list": [metric["precision"] for metric in metric_list],
            "top_k": k,
        }


if __name__ == "__main__":
    from adalflow.datasets import HotPotQA, HotPotQAData

    train_dataset = HotPotQA(split="train", size=10)
    data: HotPotQAData = train_dataset[0]
    gold_titles = data.gold_titles
    context_titles = data.context["title"]
    print(f"gold_titles: {gold_titles}, context_titles: {context_titles}")
    print(f"train: {len(train_dataset)}, example: {train_dataset[0]}")

    # compute the recall and precision for 10 items
    retriever_eval = RetrieverEvaluator()

    gt_contexts = [list(data.gold_titles) for data in train_dataset[:10]]

    retrieved_contexts = [list(data.context["title"]) for data in train_dataset[:10]]

    result = retriever_eval.compute(retrieved_contexts, gt_contexts)

    print(f"result: {result}")
